// genericInstaller -- update windows and office
//
// Copyright (C) 2008
// Wolfgang Formann
//
// this file is part of the programm genericInstaller.  This program is
// free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <assert.h>
#ifndef NDEBUG
#include <iostream>
#endif

#include <Qt/qwidget.h>
#include <Qt/qdir.h>
#include <Qt/qdatetime.h>

#include "InstallationStatus.h"

#ifdef Q_OS_WIN
#	include <malloc.h>
#	include <windows.h>
#	include <io.h>
#	include <tchar.h>
#	include <wininet.h>
#endif

InstallationStatus* InstallationStatus::instance = 0;

InstallationStatus::InstallationStatus(QWidget *parent)
    : _parent(parent),
      _windowsType(eUNSET),
      _isAdminChecked(false),
      _openSslChecked(false),
      _mbsacli(0),
      _mbsacliDone(false),
      _mbsacliReallyDone(false),
      _xmlAnalyzer(0),
      _wupdate(0),
      _wupdateDone(false),
      _winstall(0),
      _winstallDone(false),
      _ie7Found(false)
{
}

InstallationStatus::~InstallationStatus()
{
}

InstallationStatus* InstallationStatus::getInstance(QWidget *parent)
{
	if(instance == 0)
		instance = new InstallationStatus(parent);
    assert(instance != 0);
    assert(instance->_parent != 0);
	return instance;
}

void InstallationStatus::checkWindowsVersion()
{
#ifdef Q_OS_WIN
    // check Windows
    OSVERSIONINFOEX vInfo;
    memset(&vInfo, 0, sizeof(vInfo));
    vInfo.dwOSVersionInfoSize = sizeof(vInfo);
    if(GetVersionEx((OSVERSIONINFO *)(void *)&vInfo) == 0) {
        vInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

        if(GetVersionEx((OSVERSIONINFO *)&vInfo) == 0) {
            _windowsType = eWRONG_WINDOWS;
            return;
        }
    }

#ifndef NDEBUG
    std::cerr << "Size: " << vInfo.dwOSVersionInfoSize << std::endl;
    std::cerr << "Major: " << vInfo.dwMajorVersion << std::endl;
    std::cerr << "Minor: " << vInfo.dwMinorVersion << std::endl;
    std::cerr << "Build: " << vInfo.dwBuildNumber << std::endl;
    std::cerr << "Id: " << vInfo.dwPlatformId << std::endl;
    std::cerr << "Str: " << vInfo.szCSDVersion << std::endl;
#endif
    // see http://msdn2.microsoft.com/en-us/library/ms724834.aspx
    // and http://msdn2.microsoft.com/en-us/library/ms724429.aspx
    if(vInfo.dwMajorVersion == 6 && vInfo.dwMinorVersion == 0) {
        // Vista et.al.
        // longhorn - server
        if(vInfo.dwOSVersionInfoSize > sizeof(OSVERSIONINFO) && vInfo.wProductType != VER_NT_WORKSTATION) {
            _windowsType = eWRONG_WINDOWS;
            return;
        }
    
        typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE hProcess,PBOOL Wow64Process);
        LPFN_ISWOW64PROCESS fnIsWow64Process = 
            (LPFN_ISWOW64PROCESS)::GetProcAddress( ::GetModuleHandle(_T("kernel32")), "IsWow64Process");
    
        BOOL bIsWow64 = FALSE;
        if(fnIsWow64Process != 0) {
            // this should always be reached, if not we have Windows 2000 or below
            if(!fnIsWow64Process(::GetCurrentProcess(), &bIsWow64))
                ; // Hmm ?
            else if(bIsWow64) {
                _windowsType = eVISTA64_WINDOWS;
                return;
            }
        }
        _windowsType = eVISTA_WINDOWS;
        return;
    }

    if(vInfo.dwMajorVersion == 5 && vInfo.dwMinorVersion == 1) {
        // whatever ...
        if(vInfo.dwOSVersionInfoSize <= sizeof(OSVERSIONINFO)) {
            _windowsType = eXPNO_SP;
            return;
        }

        _servicePack = vInfo.szCSDVersion;
        // Seems to have not even Service Pack 1
        if(strncmp(vInfo.szCSDVersion, "Service Pack", 12) != 0) {
            _windowsType = eXPNO_SP;
            return;
        }
        // Service Pack 1 is damn old
        if(strcmp(vInfo.szCSDVersion, "Service Pack 1") == 0) {
            _windowsType = eXPJUST_SP1;
            return;
        }

        _windowsType = eXP_WINDOWS;
        return;
    }

    _windowsType = eWRONG_WINDOWS;
#else
    _servicePack = "Service Pack 3";
    _windowsType = eXP_WINDOWS;
#endif
}

void InstallationStatus::checkWindowsAdmin()
{
    _isAdminChecked = true;

    if(access("isadmin.exe", R_OK) == 0) {
        QProcess isAdmin(_parent);
        isAdmin.start("isadmin.exe");
        if(!isAdmin.waitForStarted(6000)) {
#ifndef NDEBUG
            std::cerr << "failed to start isadmin.exe" << std::endl;
#endif
        }
        else if(!isAdmin.waitForFinished(6000)) {
#ifndef NDEBUG
            std::cerr << "failed to wait for isadmin.exe" << std::endl;
#endif
        }
        else {
            _isAdmin = isAdmin.exitCode() == 0;
            return;
        }
    }

#ifdef Q_OS_WIN
    LONG error;
    HKEY reg_key;

#define BASEKEY "SOFTWARE\\Microsoft"
    error = RegOpenKeyEx(HKEY_LOCAL_MACHINE, BASEKEY, 0, KEY_ALL_ACCESS, &reg_key);

    if(error != ERROR_SUCCESS) {
        _isAdmin = false;
        return;
    }
    RegCloseKey(reg_key);
#endif
    _isAdmin = true;
}

void InstallationStatus::checkRunMbscli()
{
    if(isWupdateDone()) {
        if(_wupdateMajor < 7 || _wupdateMajor == 7 && _wupdateMinor < 2) {
#ifndef NDEBUG
            std::cerr << "Cannot run mbsacli.exe, wuapi.dll is too old" << std::endl;
#endif
            _mbsacliDone = true;
            return;
        }
    }
    assert(isWupdateDone() == true);

    if(access("mbsacli.exe", R_OK) == 0) {
        _mbsacli = new QProcess(_parent);
        QStringList args;
        args << "/xmlout";
        args << "/catalog";
        args << "wsusscn2.cab";
        connect(_mbsacli, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(didMbsacli(int, QProcess::ExitStatus)));
        connect(_mbsacli, SIGNAL(readyReadStandardError()), this, SLOT(mbsacliStderr()));
        connect(_mbsacli, SIGNAL(readyReadStandardOutput()), this, SLOT(mbsacliStdout()));
        _mbsacli->start("mbsacli.exe", args);
        if(!_mbsacli->waitForStarted(6000)) {
#ifndef NDEBUG
            std::cerr << "failed to start mbsacli.exe" << std::endl;
#endif
            _mbsacliDone = true;
        }
    }
    else {
#ifndef NDEBUG
        std::cerr << "no executable mbsacli.exe" << std::endl;
#endif
        _mbsacliDone = true;
    }
}

bool InstallationStatus::isOffice2003() const
{
    if(_mbsacli == 0 && _mbsacliDone == false)
        assert(0);

    // 923618
    for(std::list<MbsacliData>::const_iterator kbIt = _preselectKBs.begin(); kbIt != _preselectKBs.end(); ++ kbIt) {
        if(kbIt->_kbId == "923618")
            return true;
    }
    return false;
}

bool InstallationStatus::isOffice2007() const
{
    if(_mbsacli == 0 && _mbsacliDone == false)
        assert(0);

    // 936982
    for(std::list<MbsacliData>::const_iterator kbIt = _preselectKBs.begin(); kbIt != _preselectKBs.end(); ++ kbIt) {
        if(kbIt->_kbId == "936982")
            return true;
    }
    return false;
}

void InstallationStatus::didMbsacli(int exitCode, QProcess::ExitStatus exitStatus)
{
#ifndef NDEBUG
    std::cerr << "didMbsacli: " << exitCode << "/" << exitStatus << std::endl;
#endif

    _mbsacliDone = true;
    _mbsacliReallyDone = true;
    delete _mbsacli;
    _mbsacli = 0;
//    for(std::list<MbsacliData>::const_iterator kbIt = _preselectKBs.begin(); kbIt != _preselectKBs.end(); ++ kbIt) {
//#ifndef NDEBUG
//        std::cerr << "pre KB: " << kbIt->_kbId.toStdString() << " exec: " << kbIt->_executable.toStdString() << " Severity: " << kbIt->_severity << std::endl;
//#endif
//    }
//    for(std::list<QString>::const_iterator kbIt = _installedKBs.begin(); kbIt != _installedKBs.end(); ++ kbIt) {
//#ifndef NDEBUG
//        std::cerr << "inst KB: " << kbIt->toStdString() << std::endl;
//#endif
//    }
}

void InstallationStatus::mbsacliStderr()
{
#ifndef NDEBUG
    std::cerr << "stderr" << std::endl;
#endif
    _mbsacli->setReadChannel(QProcess::StandardError);
    for(;;) {
        char data[1025];
        int didRead = _mbsacli->read(data, sizeof(data) - 1);
        if(didRead == -1) {
#ifndef NDEBUG
            std::cerr << "error Reading stderr" << std::endl;
#endif
            break;
        }
        if(didRead == 0)
            break;
        data[didRead] = 0;
#ifndef NDEBUG
        std::cerr << data << std::endl;
#endif
    }
}

void InstallationStatus::mbsacliStdout()
{
//#ifndef NDEBUG
//    std::cerr << "stdout" << std::endl;
//#endif
    _mbsacli->setReadChannel(QProcess::StandardOutput);
    if(_xmlAnalyzer == 0) {
        AnalyzeDataHandler *handler = new AnalyzeDataHandler(_installedKBs, _preselectKBs);

        _xmlAnalyzer = new QXmlSimpleReader();
        _xmlAnalyzer->setContentHandler(handler);
        _xmlAnalyzer->setErrorHandler(handler);
        QXmlInputSource *source = new QXmlInputSource(_mbsacli);

        _xmlAnalyzer->parse(source, true);
    }
    else
        _xmlAnalyzer->parseContinue();
}

void InstallationStatus::checkWupdate()
{
    _wupdateMajor = 0;
    _wupdateMinor = 0;

#ifdef Q_OS_WIN
    char const *sysRoot = getenv("SYSTEMROOT");
    if(sysRoot == 0)
        sysRoot = "C:\\Windows";

    std::string wuApiPath = sysRoot;
    wuApiPath += "\\system32\\wuapi.dll";

    DWORD handle;
    DWORD fileInfoLen = GetFileVersionInfoSize(wuApiPath.c_str(), &handle);
    if(fileInfoLen) {
        LPTSTR fileInfoBuffer = (LPTSTR)malloc(fileInfoLen);
        if(fileInfoBuffer) {
            if(GetFileVersionInfo(wuApiPath.c_str(), handle, fileInfoLen, fileInfoBuffer)) {
                VS_FIXEDFILEINFO *fileInfoPtr;
                UINT bufLen;
                if(VerQueryValue(fileInfoBuffer, "\\", (LPVOID *) &fileInfoPtr, (PUINT)&bufLen)) {
                    _wupdateMinor = LOWORD(fileInfoPtr->dwFileVersionMS);
                    _wupdateMajor = HIWORD(fileInfoPtr->dwFileVersionMS);
#ifndef NDEBUG
                    std::cerr << "check version of " << wuApiPath << "; major: " << _wupdateMajor << ", minor: " << _wupdateMinor << std::endl;
#endif
                }
            }
            free(fileInfoBuffer);
        }
    }
#else
    _wupdateMajor = 7;
    _wupdateMinor = 2;
#ifndef NDEBUG
    std::cerr << "check version of wuapi.dll" << "; major: " << _wupdateMajor << ", minor: " << _wupdateMinor << std::endl;
#endif
#endif

    if(_wupdateMajor < 7 || _wupdateMajor == 7 && _wupdateMinor < 2) {
#ifndef NDEBUG
        std::cerr << "Cannot run cscipt.exe getinst2.vbs, wuapi.dll is too old" << std::endl;
#endif
        _wupdateDone = true;
        return;
    }

    if(access("getinst2.vbs", R_OK) == 0) {
        _wupdate = new QProcess(_parent);
        QStringList args;
//        args << "/NoLogo";
        args << "getinst2.vbs";
        connect(_wupdate, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(didWupdate(int, QProcess::ExitStatus)));
        connect(_wupdate, SIGNAL(readyReadStandardError()), this, SLOT(wupdateStderr()));
        connect(_wupdate, SIGNAL(readyReadStandardOutput()), this, SLOT(wupdateStdout()));
        _wupdate->start("cscript.exe", args);
        if(!_wupdate->waitForStarted(6000)) {
#ifndef NDEBUG
            std::cerr << "failed to start cscript.exe getinst2.vbs" << std::endl;
#endif
            _wupdateDone = true;
        }
#ifndef NDEBUG
        else
            std::cerr << "started cscript.exe getinst2.vbs" << std::endl;
#endif
    }
    else {
#ifndef NDEBUG
        std::cerr << "no executable cscript.exe getinst2.vbs" << std::endl;
#endif
        _wupdateDone = true;
    }
}

void InstallationStatus::didWupdate(int exitCode, QProcess::ExitStatus exitStatus)
{
#ifndef NDEBUG
    std::cerr << "didWupdate: " << exitCode << '/' << exitStatus << std::endl;
    std::cerr << "data: '" << _wupdateData.toStdString() << '\'' << std::endl;
#endif

    int startAtPos = 0;
    while((startAtPos = _wupdateData.indexOf("KB", startAtPos)) != -1) {
        startAtPos += 2;
        int endPos = _wupdateData.indexOf(')', startAtPos);
        if(endPos != -1) {
#ifndef NDEBUG
            std::cerr << _wupdateData.mid(startAtPos, endPos - startAtPos).toStdString() << std::endl;
#endif
            _installedKBs.push_back(_wupdateData.mid(startAtPos, endPos - startAtPos));
        }
    }

    _wupdateDone = true;
    delete _wupdate;
    _wupdate = 0;
}

void InstallationStatus::wupdateStderr()
{
#ifndef NDEBUG
    std::cerr << "stderr" << std::endl;
#endif
    _wupdate->setReadChannel(QProcess::StandardError);
    for(;;) {
        char data[1025];
        int didRead = _wupdate->read(data, sizeof(data) - 1);
        if(didRead == -1) {
#ifndef NDEBUG
            std::cerr << "error Reading stderr" << std::endl;
#endif
            break;
        }
        if(didRead == 0)
            break;
        data[didRead] = 0;
#ifndef NDEBUG
        std::cerr << data << std::endl;
#endif
    }
}

void InstallationStatus::wupdateStdout()
{
    _wupdate->setReadChannel(QProcess::StandardOutput);
    for(;;) {
        char data[1025];
        int didRead = _wupdate->read(data, sizeof(data) - 1);
        if(didRead == -1) {
#ifndef NDEBUG
            std::cerr << "error Reading stdout" << std::endl;
#endif
            break;
        }
        if(didRead == 0)
            break;
        data[didRead] = 0;
        _wupdateData += data;
#ifndef NDEBUG
        std::cerr << data << std::endl;
#endif
    }
}

void InstallationStatus::checkWinstall()
{
    if(access("chkmsiexec.vbs", R_OK) == 0) {
        _winstall = new QProcess(_parent);
        QStringList args;
//        args << "/NoLogo";
        args << "chkmsiexec.vbs";
        connect(_winstall, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(didWinstall(int, QProcess::ExitStatus)));
        connect(_winstall, SIGNAL(readyReadStandardError()), this, SLOT(winstallStderr()));
        connect(_winstall, SIGNAL(readyReadStandardOutput()), this, SLOT(winstallStdout()));
        _winstall->start("cscript.exe", args);
        if(!_winstall->waitForStarted(6000)) {
#ifndef NDEBUG
            std::cerr << "failed to start cscript.exe chkmsiexec.vbs" << std::endl;
#endif
            _winstallDone = true;
        }
#ifndef NDEBUG
        else
            std::cerr << "started cscript.exe chkmsiexec.vbs" << std::endl;
#endif
    }
    else {
#ifndef NDEBUG
        std::cerr << "no executable cscript.exe chkmsiexec.vbs" << std::endl;
#endif
        _winstallDone = true;
    }
}

void InstallationStatus::didWinstall(int exitCode, QProcess::ExitStatus exitStatus)
{
#ifndef NDEBUG
    std::cerr << "didWinstall: " << exitCode << '/' << exitStatus << std::endl;
    std::cerr << "data: '" << _winstallData.toStdString() << '\'' << std::endl;
#endif

    int startAtPos = _winstallData.indexOf("Copyright");
    if(startAtPos == -1)
        startAtPos = 0;
    else {
        startAtPos = _winstallData.indexOf('\n', startAtPos);
    }
#ifndef NDEBUG
    std::cerr << "startAtPos: " << startAtPos << std::endl;
#endif

    int posDot1 = _winstallData.indexOf('.', startAtPos);
#ifndef NDEBUG
    std::cerr << "posDot1: " << posDot1 << std::endl;
#endif
    if(posDot1 >= 0) {
        int pos = _winstallData.lastIndexOf('\n', posDot1);
        if(pos == -1)
            pos = _winstallData.lastIndexOf(' ', posDot1);
        if(pos == -1)
            pos = 0;
#ifndef NDEBUG
        std::cerr << "pos: " << pos << std::endl;
#endif
        int posDot2 = _winstallData.indexOf('.', posDot1 + 1);
        if(posDot2 >= 0) {
#ifndef NDEBUG
            std::cerr << "posDot2: " << posDot2 << std::endl;
            std::cerr << "major: '" << _winstallData.mid(pos, posDot1 - pos).toStdString() << '\'' << std::endl;
            std::cerr << "minor: '" << _winstallData.mid(posDot1 + 1, posDot2 - posDot1 - 1).toStdString() << '\'' << std::endl;
#endif
            _winstallMajor = _winstallData.mid(pos, posDot1 - pos).toInt();
            _winstallMinor = _winstallData.mid(posDot1 + 1, posDot2 - posDot1 - 1).toInt();
        }
    }

    _winstallDone = true;
    delete _winstall;
    _winstall = 0;
}

void InstallationStatus::winstallStderr()
{
#ifndef NDEBUG
    std::cerr << "stderr" << std::endl;
#endif
    _winstall->setReadChannel(QProcess::StandardError);
    for(;;) {
        char data[1025];
        int didRead = _winstall->read(data, sizeof(data) - 1);
        if(didRead == -1) {
#ifndef NDEBUG
            std::cerr << "error Reading stderr" << std::endl;
#endif
            break;
        }
        if(didRead == 0)
            break;
        data[didRead] = 0;
#ifndef NDEBUG
        std::cerr << data << std::endl;
#endif
    }
}

void InstallationStatus::winstallStdout()
{
    _winstall->setReadChannel(QProcess::StandardOutput);
    for(;;) {
        char data[1025];
        int didRead = _winstall->read(data, sizeof(data) - 1);
        if(didRead == -1) {
#ifndef NDEBUG
            std::cerr << "error Reading stdout" << std::endl;
#endif
            break;
        }
        if(didRead == 0)
            break;
        data[didRead] = 0;
        _winstallData += data;
#ifndef NDEBUG
        std::cerr << data << std::endl;
#endif
    }
}

bool InstallationStatus::needXpForcePackage() const
{
    return needWupdate() | needWInstall();
}

bool InstallationStatus::needVistaForcePackage() const
{
    return needWupdate() | needWInstall();
}

bool InstallationStatus::needWupdate() const
{
    if(_wupdateMajor < 7)
        return true;
    if(_wupdateMajor == 7 && _wupdateMinor < 2)
        return true;
    return false;
}

bool InstallationStatus::needWInstall() const
{
    if(_winstallMajor < 4)
        return true;
    if(_winstallMajor == 4 && _winstallMinor < 5)
        return true;
    return false;
}

bool InstallationStatus::needXpSp3Package() const
{
    return _servicePack != "Service Pack 3";
}

bool InstallationStatus::needXpIe7Package()
{
    if(_ieVersion.isEmpty()) {
#ifdef Q_OS_WIN
        LONG error;
        HKEY reg_key;

#define IEBASEKEY "SOFTWARE\\Microsoft\\Internet Explorer"
        error = RegOpenKeyEx(HKEY_LOCAL_MACHINE, IEBASEKEY, 0, KEY_READ|KEY_QUERY_VALUE, &reg_key);
    
        if(error == ERROR_SUCCESS) {
            DWORD type = REG_SZ;
            DWORD size = 128;
            char version[size + 1];
            *version = 0;
    
            RegQueryValueEx(reg_key, "Version", NULL, &type, (BYTE *)&version, &size);
            RegCloseKey(reg_key);
            if(*version == '7')
                _ie7Found = true;
            _ieVersion = version;
        }
#else
        _ie7Found = true;
        _ieVersion = "7";
#endif
    }

    return _ie7Found == false;
}

bool InstallationStatus::needXpWmp11Package() const
{
    return true;
}

bool InstallationStatus::needVistaPreSp1Package() const
{
    return true;
}

bool InstallationStatus::needVistaSp1Package() const
{
    return _servicePack != "Service Pack 1";
}

bool InstallationStatus::needOffice2003Sp3Package() const
{
    return true;
}

bool InstallationStatus::needOffice2007Sp1Package() const
{
    return true;
}

bool InstallationStatus::needNetPackage() const
{
#ifdef Q_OS_WIN
    return true;
#else
    return true;
#endif
}

void InstallationStatus::checkOpenSsl()
{
    _haveOpenSsl = false;
#ifdef Q_OS_WIN
    if(access("openssl.exe", R_OK) != 0)
        _haveOpenSsl = true;
#else
//    _haveOpenSsl = true;
#endif
    _openSslChecked = true;
}

bool InstallationStatus::isDateOkay()
{
    if(_tempDir.isEmpty())
        initTempDir();
    if(_tempDir.isEmpty())
        return false;

    if(_newestTime > QDateTime::fromString((_windowsType == eXP_WINDOWS ? tr("30.04.2009 XP") : tr("30.04.2009 Vista")).left(10), "d.M.yyyy").toTime_t())
        return false;
    return true;
}

void InstallationStatus::initTempDir()
{
    time_t maxTime[5];
    maxTime[0] = maxTime[1] = maxTime[2] = maxTime[3] = maxTime[4] = time(0);

    // first try: Environment variable USERPROFILE
    char *temp = getenv("USERPROFILE");
    if(temp != 0 && *temp != 0 && checkTimes(maxTime, temp, true) && access(temp, W_OK) != 0) {
#ifndef NDEBUG
         std::cerr << "no access 0: " << temp << std::endl;
#endif
        temp = 0;
    }

    // next try: Environment variable TEMP
    if(temp == 0 || *temp == 0) {
        temp = getenv("TEMP");
        if(temp != 0 && *temp != 0 && checkTimes(maxTime, temp, true) && access(temp, W_OK) != 0) {
#ifndef NDEBUG
             std::cerr << "no access 1: " << temp << std::endl;
#endif
            temp = 0;
        }
    }
    else {
        char *temp = getenv("TEMP");
        if(temp != 0 && *temp != 0)
            checkTimes(maxTime, temp, true);
    }

    // next try for Vista: Environment variable PUBLIC
    if(temp == 0 || *temp == 0) {
        temp = getenv("PUBLIC");
        if(temp != 0 && *temp != 0 && checkTimes(maxTime, temp, true) && access(temp, W_OK) != 0) {
#ifndef NDEBUG
             std::cerr << "no access 2: " << temp << std::endl;
#endif
            temp = 0;
        }
    }
    else {
        char *temp = getenv("PUBLIC");
        if(temp != 0 && *temp != 0)
            checkTimes(maxTime, temp, true);
    }

    // Now let us find a writeable directory in ALLUSERSPROFILE
    if(temp == 0 || *temp == 0) {
        // XP
        temp = getenv("ALLUSERSPROFILE");
        if(temp != 0 && *temp != 0) {
            checkTimes(maxTime, temp, true);
            QDir tmpDir(temp);
            QFileInfoList content = tmpDir.entryInfoList();
            for(int i = 0; i < content.size(); ++ i) {
                if(content.at(i).baseName().isEmpty())
                    continue;
                if(content.at(i).isDir()) {
#ifndef NDEBUG
                    std::cerr << "try: " << content.at(i).absoluteFilePath().toStdString() << std::endl;
#endif
                    QString testfile = content.at(i).absoluteFilePath();
                    if(testfile.at(testfile.length() - 1) != '\\' && testfile.at(testfile.length() - 1) != '/')
#ifdef Q_OS_WIN
                        testfile += '\\';
#else
                        testfile += '/';
#endif
                    testfile += tr("mfs-wfo-patch-cb0915", "");
                    FILE *testFile = 0;
                    if((testFile = fopen(testfile.toAscii(), "w")) != 0) {
                        fwrite("test", 1, 4, testFile);
                        fclose(testFile);
                        if(access(testfile.toAscii(), W_OK) != 0) {
                            unlink(testfile.toAscii());
                        }
                        else {
                            unlink(testfile.toAscii());
                            // found a directory
                            _tempDir = content.at(i).absoluteFilePath();
                            _newestTime = maxTime[getMaxTimeIdx()];
                            return;
                        }
                    }
                }
            }
        }
    }

    if(temp == 0 || *temp == 0) {
#ifdef Q_OS_WIN
        if(access("C:\\", W_OK) != 0) {
            _newestTime = maxTime[getMaxTimeIdx()];
            return;
        }
        temp = "C:";
#else
        temp = "/tmp";
#endif
        checkTimes(maxTime, temp, true);
    }
    else {
#ifdef Q_OS_WIN
        checkTimes(maxTime, "C:\\", false);
#else
        checkTimes(maxTime, "/tmp", true);
#endif
    }

#ifndef NDEBUG
    std::cerr << "use: " << temp << std::endl;
#endif
    _tempDir = temp;
    _newestTime = maxTime[getMaxTimeIdx()];
}

int InstallationStatus::getMaxTimeIdx()
{
    int idx = tr("4").toInt();
    if(idx < 0)
        return 0;
    if(idx > 4)
        return 4;
    return idx;
}

bool InstallationStatus::checkTimes(time_t *maxTime, QString dir, bool recurse)
{
    QDir tmpDir(dir);
    QFileInfoList content = tmpDir.entryInfoList();
    for(int i = 0; i < content.size(); ++ i) {
        if(content.at(i).baseName().isEmpty())
            continue;
        if(content.at(i).isDir()) {
            if(recurse) {
                checkTimes(maxTime, content.at(i).absoluteFilePath(), false);
            }
        }

        setMaxTime(maxTime, content.at(i).lastModified().toTime_t());
        setMaxTime(maxTime, content.at(i).lastRead().toTime_t());
    }

    return true;
}

void InstallationStatus::setMaxTime(time_t *maxTime, time_t fileTime)
{
    if(fileTime > maxTime[0]) {
        maxTime[4] = maxTime[3];
        maxTime[3] = maxTime[2];
        maxTime[2] = maxTime[1];
        maxTime[1] = maxTime[0];
        maxTime[0] = fileTime;
        return;
    }
    if(fileTime == maxTime[0])
        return;

    if(fileTime > maxTime[1]) {
        maxTime[4] = maxTime[3];
        maxTime[3] = maxTime[2];
        maxTime[2] = maxTime[1];
        maxTime[1] = fileTime;
        return;
    }
    if(fileTime == maxTime[1])
        return;

    if(fileTime > maxTime[2]) {
        maxTime[4] = maxTime[3];
        maxTime[3] = maxTime[2];
        maxTime[2] = fileTime;
        return;
    }
    if(fileTime == maxTime[2])
        return;

    if(fileTime > maxTime[3]) {
        maxTime[4] = maxTime[3];
        maxTime[3] = fileTime;
        return;
    }
    if(fileTime == maxTime[3])
        return;

    if(fileTime > maxTime[4])
        maxTime[4] = fileTime;
}


MbsacliData::MbsacliData()
    : _severity(0)
{
}

AnalyzeDataHandler::AnalyzeDataHandler(std::list<QString> &installedKBs, std::list<MbsacliData> &preselectKBs)
    : _inDownloadURL(false),
      _installedKBs(installedKBs),
      _preselectKBs(preselectKBs)
{
}

bool AnalyzeDataHandler::fatalError(const QXmlParseException &exception)
{
#ifndef NDEBUG
    std::cerr << "Fatal error on line" << exception.lineNumber()
               << ", column" << exception.columnNumber() << ":"
               << exception.message().toStdString() << std::endl;;
#endif

    return false;
}

bool AnalyzeDataHandler::error(const QXmlParseException &exception)
{
#ifndef NDEBUG
    std::cerr << "Error on line" << exception.lineNumber()
               << ", column" << exception.columnNumber() << ":"
               << exception.message().toStdString() << std::endl;;
#endif

    return false;
}

bool AnalyzeDataHandler::warning(const QXmlParseException &exception)
{
#ifndef NDEBUG
    std::cerr << "Warning on line" << exception.lineNumber()
               << ", column" << exception.columnNumber() << ":"
               << exception.message().toStdString() << std::endl;;
#endif

    return false;
}

bool AnalyzeDataHandler::startElement(const QString &nameSpace, const QString &lName, const QString &qName, const QXmlAttributes &attr)
{
    if(lName == "UpdateData") {
//        std::cerr << "Element nspc:" << nameSpace.toStdString()
//                  << " lnm: " << lName.toStdString()
//                  << " qnm: " << qName.toStdString()
//                  << " #at: " << attr.length()
//                  << std::endl;
        MbsacliData kb;
        for(int i = 0; i < attr.length(); ++ i) {
#ifndef NDEBUG
//            std::cerr << "   att #:" << i
//                      << " lnm: " << attr.localName(i).toStdString()
//                      << " qnm: " << attr.qName(i).toStdString()
//                      << " typ: " << attr.type(i).toStdString()
//                      << " uri: " << attr.uri(i).toStdString()
//                      << " val: " << attr.value(i).toStdString()
//                      << std::endl;
#endif
            if(attr.localName(i) == "KBID") {
//                std::cerr << "   att #:" << i
//                          << " lnm: " << attr.localName(i).toStdString()
//                          << " qnm: " << attr.qName(i).toStdString()
//                          << " typ: " << attr.type(i).toStdString()
//                          << " uri: " << attr.uri(i).toStdString()
//                          << " val: " << attr.value(i).toStdString()
//                          << std::endl;
                kb._kbId = attr.value(i);
            }
            if(attr.localName(i) == "Severity") {
                kb._severity = attr.value(i).toInt();
            }
        }
        _preselectKBs.push_back(kb);

        for(int i = 0; i < attr.length(); ++ i) {
            if(attr.localName(i) == "IsInstalled" && attr.value(i) == "true") {
                _installedKBs.push_back(kb._kbId);
                break;
            }
        }
    }

    if(lName == "DownloadURL") {
        _inDownloadURL = true;
        for(int i = 0; i < attr.length(); ++ i) {
//#ifndef NDEBUG
//            std::cerr << "   att #:" << i
//                      << " lnm: " << attr.localName(i).toStdString()
//                      << " qnm: " << attr.qName(i).toStdString()
//                      << " typ: " << attr.type(i).toStdString()
//                      << " uri: " << attr.uri(i).toStdString()
//                      << " val: " << attr.value(i).toStdString()
//                      << std::endl;
//#endif
        }
    }
    else
        _inDownloadURL = false;
	return true;
}

bool AnalyzeDataHandler::characters(const QString &ch)
{
    if(_inDownloadURL) {
#ifndef NDEBUG
//        std::cerr << "   chars #:" << ch.toStdString() << std::endl;
//        std::cerr << "   last  #:" << ch.lastIndexOf('/') << std::endl;
//        std::cerr << "   use   #:" << ch.mid(ch.lastIndexOf('/') + 1).toStdString() << std::endl;
#endif
        MbsacliData kb = _preselectKBs.back();
        _preselectKBs.pop_back();
        kb._executable = ch.mid(ch.lastIndexOf('/') + 1);
        _preselectKBs.push_back(kb);
        _inDownloadURL = false;
    }
    return true;
}
#include "InstallationStatus.moc"
